@@ -3,32 +3,51 @@ import logging
 import os
 import shutil
 import tarfile
+import tempfile
 import time
+import uuid
 import zipfile
 from collections import deque
 from functools import wraps
 from io import BytesIO
 from pathlib import Path
 from threading import Lock
-from typing import Union, Optional
+from typing import Optional, Union
 
-from httpx import RequestError, Response, AsyncClient
 import orjson as json
+from httpx import AsyncClient, RequestError, Response
 from stamina import retry
 from tqdm import tqdm
 
-from edgar.core import text_extensions, get_edgar_data_directory
-from edgar.httpclient import http_client, async_http_client
+from edgar.core import get_edgar_data_directory, text_extensions
+from edgar.httpclient import async_http_client, http_client
 
-__all__ = ["get_with_retry", "get_with_retry_async", "stream_with_retry", "post_with_retry", "post_with_retry_async",
-           "download_file", "download_file_async", "download_json", "download_json_async", "stream_file",
-           "download_text", "download_text_between_tags", "download_bulk_data", "download_datafile",
-           "throttle_requests"]
+"""
+This module provides functions to handle HTTP requests with retry logic, throttling, and identity management.
+"""
+__all__ = [
+    "get_with_retry",
+    "get_with_retry_async",
+    "stream_with_retry",
+    "post_with_retry",
+    "post_with_retry_async",
+    "download_file",
+    "download_file_async",
+    "download_json",
+    "download_json_async",
+    "stream_file",
+    "download_text",
+    "download_text_between_tags",
+    "download_bulk_data",
+    "download_datafile",
+    "throttle_requests",
+]
 
 attempts = 6
 retry_timeout = 40
 wait_initial = 0.1
 max_requests_per_second = 8
+throttle_disabled = False
 
 
 class TooManyRequestsError(Exception):
@@ -76,8 +95,11 @@ class Throttler:
             current_time = time.monotonic()
 
             # Remove timestamps older than the time window
-            while self.request_timestamps and self.request_timestamps[
-                0] <= current_time - self.request_rate.time_window:
+            while (
+                    self.request_timestamps
+                    and self.request_timestamps[0]
+                    <= current_time - self.request_rate.time_window
+            ):
                 self.request_timestamps.popleft()
 
             if len(self.request_timestamps) < self.request_rate.max_requests:
@@ -92,7 +114,9 @@ class Throttler:
 
     def update_metrics(self):
         self.total_calls += 1
-        current_call_rate: float = len(self.request_timestamps) / self.request_rate.time_window
+        current_call_rate: float = (
+                len(self.request_timestamps) / self.request_rate.time_window
+        )
         self.peak_call_rate = max(self.peak_call_rate, current_call_rate)
 
     def get_metrics(self):
@@ -105,7 +129,9 @@ class Throttler:
 
     def print_metrics(self):
         metrics = self.get_metrics()
-        print(f"Metrics for decorated functions: {', '.join(metrics['decorated_functions'])}")
+        print(
+            f"Metrics for decorated functions: {', '.join(metrics['decorated_functions'])}"
+        )
         print(f"Total calls: {metrics['total_calls']}")
         print(f"Peak call rate: {metrics['peak_call_rate']:.2f} calls per second")
 
@@ -117,6 +143,7 @@ def throttle_requests(request_rate=None, requests_per_second=None, **kwargs):
     """
     Decorator to throttle the number of requests per second.
     """
+
     if requests_per_second is not None:
         request_rate = RequestRate(max_requests=requests_per_second, time_window=1)
     elif request_rate is None:
@@ -133,10 +160,13 @@ def throttle_requests(request_rate=None, requests_per_second=None, **kwargs):
     def decorator(func):
         @wraps(func)
         def wrapper(*args, **kwargs):
-            throttler.wait_for_ticket()
-            result = func(*args, **kwargs)
-            throttler.update_metrics()
-            return result
+            if throttle_disabled:
+                return func(*args, **kwargs)
+            else:
+                throttler.wait_for_ticket()
+                result = func(*args, **kwargs)
+                throttler.update_metrics()
+                return result
 
         # Store the decorated function name
         if throttler.decorated_functions is None:
@@ -167,7 +197,9 @@ def with_identity(func):
         headers["User-Agent"] = identity
         kwargs["headers"] = headers
 
-        return func(url, identity=identity, identity_callable=identity_callable, *args, **kwargs)
+        return func(
+            url, identity=identity, identity_callable=identity_callable, *args, **kwargs
+        )
 
     return wrapper
 
@@ -187,13 +219,21 @@ def async_with_identity(func):
         headers["User-Agent"] = identity
         kwargs["headers"] = headers
 
-        return func(client, url, identity=identity, identity_callable=identity_callable, *args, **kwargs)
+        return func(
+            client,
+            url,
+            identity=identity,
+            identity_callable=identity_callable,
+            *args,
+            **kwargs,
+        )
 
     return wrapper
 
 
-
-@retry(on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
+@retry(
+    on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial
+)
 @with_identity
 @throttle_requests(requests_per_second=max_requests_per_second)
 def get_with_retry(url, identity=None, identity_callable=None, **kwargs):
@@ -217,15 +257,23 @@ def get_with_retry(url, identity=None, identity_callable=None, **kwargs):
         if response.status_code == 429:
             raise TooManyRequestsError(url)
         elif is_redirect(response):
-            return get_with_retry(url=response.headers["Location"], identity=identity, identity_callable=identity_callable,
-                                 **kwargs)
+            return get_with_retry(
+                url=response.headers["Location"],
+                identity=identity,
+                identity_callable=identity_callable,
+                **kwargs,
+            )
         return response
 
 
-@retry(on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
+@retry(
+    on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial
+)
 @async_with_identity
 @throttle_requests(requests_per_second=max_requests_per_second)
-async def get_with_retry_async(client: AsyncClient, url, identity=None, identity_callable=None, **kwargs):
+async def get_with_retry_async(
+        client: AsyncClient, url, identity=None, identity_callable=None, **kwargs
+):
     """
     Sends an asynchronous GET request with retry functionality and identity handling.
 
@@ -245,12 +293,19 @@ async def get_with_retry_async(client: AsyncClient, url, identity=None, identity
     if response.status_code == 429:
         raise TooManyRequestsError(url)
     elif is_redirect(response):
-        return await get_with_retry_async(client=client, url=response.headers["Location"], identity=identity,
-                                            identity_callable=identity_callable, **kwargs)
+        return await get_with_retry_async(
+            client=client,
+            url=response.headers["Location"],
+            identity=identity,
+            identity_callable=identity_callable,
+            **kwargs,
+        )
     return response
 
 
-@retry(on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
+@retry(
+    on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial
+)
 @with_identity
 @throttle_requests(requests_per_second=max_requests_per_second)
 def stream_with_retry(url, identity=None, identity_callable=None, **kwargs):
@@ -274,19 +329,25 @@ def stream_with_retry(url, identity=None, identity_callable=None, **kwargs):
             if response.status_code == 429:
                 raise TooManyRequestsError(url)
             elif is_redirect(response):
-                response = stream_with_retry(response.headers["Location"],
-                                        identity=identity,
-                                        identity_callable=identity_callable, **kwargs)
+                response = stream_with_retry(
+                    response.headers["Location"],
+                    identity=identity,
+                    identity_callable=identity_callable,
+                    **kwargs,
+                )
                 yield from response
             else:
                 yield response
 
 
-@retry(on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
+@retry(
+    on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial
+)
 @with_identity
 @throttle_requests(requests_per_second=max_requests_per_second)
-def post_with_retry(url, data=None, json=None, identity=None, identity_callable=None,
-                    **kwargs):
+def post_with_retry(
+        url, data=None, json=None, identity=None, identity_callable=None, **kwargs
+):
     """
     Sends a POST request with retry functionality and identity handling.
 
@@ -309,20 +370,31 @@ def post_with_retry(url, data=None, json=None, identity=None, identity_callable=
         if response.status_code == 429:
             raise TooManyRequestsError(url)
         elif is_redirect(response):
-            return post_with_retry(response.headers["Location"], data=data, json=json, identity=identity,
-                                   identity_callable=identity_callable, **kwargs)
+            return post_with_retry(
+                response.headers["Location"],
+                data=data,
+                json=json,
+                identity=identity,
+                identity_callable=identity_callable,
+                **kwargs,
+            )
         return response
 
 
-@retry(on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
+@retry(
+    on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial
+)
 @async_with_identity
 @throttle_requests(requests_per_second=max_requests_per_second)
-async def post_with_retry_async(client: AsyncClient, 
-                                url,
-                                data=None,
-                                json=None,
-                                identity=None,
-                                identity_callable=None, **kwargs):
+async def post_with_retry_async(
+        client: AsyncClient,
+        url,
+        data=None,
+        json=None,
+        identity=None,
+        identity_callable=None,
+        **kwargs,
+):
     """
     Sends an asynchronous POST request with retry functionality and identity handling.
 
@@ -344,8 +416,15 @@ async def post_with_retry_async(client: AsyncClient,
     if response.status_code == 429:
         raise TooManyRequestsError(url)
     elif is_redirect(response):
-        return await post_with_retry_async(client, response.headers["Location"], data=data, json=json, identity=identity,
-                                            identity_callable=identity_callable, **kwargs)
+        return await post_with_retry_async(
+            client,
+            response.headers["Location"],
+            data=data,
+            json=json,
+            identity=identity,
+            identity_callable=identity_callable,
+            **kwargs,
+        )
     return response
 
 
@@ -362,13 +441,14 @@ def decode_content(content: bytes) -> str:
     Decode the content of a file.
     """
     try:
-        return content.decode('utf-8')
+        return content.decode("utf-8")
     except UnicodeDecodeError:
-        return content.decode('latin-1')
+        return content.decode("latin-1")
 
 
-def save_or_return_content(content: Union[str, bytes], path: Optional[Union[str, Path]]) -> Union[
-    str, bytes, None]:
+def save_or_return_content(
+        content: Union[str, bytes], path: Optional[Union[str, Path]]
+) -> Union[str, bytes, None]:
     """
     Save the content to a specified path or return the content.
 
@@ -400,7 +480,9 @@ def save_or_return_content(content: Union[str, bytes], path: Optional[Union[str,
     return content
 
 
-def download_file(url: str, as_text: bool = None, path: Optional[Union[str, Path]] = None) -> Union[str, bytes, None]:
+def download_file(
+        url: str, as_text: bool = None, path: Optional[Union[str, Path]] = None
+) -> Union[str, bytes, None]:
     """
     Download a file from a URL.
 
@@ -427,7 +509,7 @@ def download_file(url: str, as_text: bool = None, path: Optional[Union[str, Path
     # Check if the content is gzip-compressed
     if url.endswith("gz"):
         binary_file = BytesIO(response.content)
-        with gzip.open(binary_file, 'rb') as f:
+        with gzip.open(binary_file, "rb") as f:
             file_content = f.read()
             if as_text:
                 file_content = decode_content(file_content)
@@ -445,8 +527,12 @@ def download_file(url: str, as_text: bool = None, path: Optional[Union[str, Path
     return save_or_return_content(file_content, path)
 
 
-async def download_file_async(client: AsyncClient, url: str, as_text: bool = None, path: Optional[Union[str, Path]] = None) -> Union[
-    str, bytes, None]:
+async def download_file_async(
+        client: AsyncClient,
+        url: str,
+        as_text: bool = None,
+        path: Optional[Union[str, Path]] = None,
+) -> Union[str, bytes, None]:
     """
     Download a file from a URL asynchronously.
 
@@ -483,77 +569,133 @@ async def download_file_async(client: AsyncClient, url: str, as_text: bool = Non
     return save_or_return_content(content, path)
 
 
-CHUNK_SIZE = 4 * 1024 * 1024  # 4MB
+# Optimized chunk sizes for different file sizes
+CHUNK_SIZE_SMALL = 1 * 1024 * 1024      # 1MB for files < 100MB
+CHUNK_SIZE_MEDIUM = 8 * 1024 * 1024     # 8MB for files 100MB - 500MB  
+CHUNK_SIZE_LARGE = 32 * 1024 * 1024     # 32MB for files > 500MB
+CHUNK_SIZE_DEFAULT = 4 * 1024 * 1024    # 4MB default (backward compatibility)
 
 
-@retry(on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
+@retry(
+    on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial
+)
 @with_identity
 @throttle_requests(requests_per_second=max_requests_per_second)
-async def stream_file(url: str,
-                      as_text: bool = None,
-                      path: Optional[Union[str, Path]] = None,
-                      client: Optional[AsyncClient]=None,
-                      **kwargs) -> Union[
-    str, bytes, None]:
+async def stream_file(
+        url: str,
+        path: Optional[Union[str, Path]] = None,
+        client: Optional[AsyncClient] = None,
+        **kwargs,
+) -> Union[str, bytes, None]:
     """
     Download a file from a URL asynchronously with progress bar using httpx.
+    
+    Always streams to disk to avoid memory accumulation, then provides atomic 
+    file operations for reliability with large files.
 
     Args:
         url (str): The URL of the file to download.
-        as_text (bool, optional): Whether to download the file as text or binary.
+        as_text (bool, optional): Whether to return the file content as text or binary.
             If None, the default is determined based on the file extension. Defaults to None.
         path (str or Path, optional): The path where the file should be saved.
+            If None, returns the content instead of saving.
         client: The httpx.AsyncClient instance
 
     Returns:
-        str or bytes: The content of the downloaded file, either as text or binary data.
+        str or bytes or None: The content of the downloaded file if path is None,
+        otherwise None after saving to path.
     """
-    if as_text is None:
-        # Set the default based on the file extension
-        as_text = url.endswith(text_extensions)
 
-    async with async_http_client(client) as async_client:
-        async with async_client.stream('GET', url) as response:
-            inspect_response(response)
-            total_size = int(response.headers.get('Content-Length', 0))
+    # Create temporary directory for atomic downloads
+    temp_dir = tempfile.mkdtemp(prefix="edgar_")
+    temp_file = Path(temp_dir) / f"download_{uuid.uuid1()}"
 
-            if as_text:
-                # Download as text
-                content = await response.text()
-                return content
-            else:
-                # Download as binary
-                content = b''
+    try:
+        async with async_http_client(client) as async_client:
+            async with async_client.stream("GET", url) as response:
+                inspect_response(response)
+                total_size = int(response.headers.get("Content-Length", 0))
+
+                # Setup progress bar
+                total_mb = total_size / (1024 * 1024) if total_size > 0 else None
                 progress_bar = tqdm(
-                    total=total_size / (1024 * 1024),
-                    unit='MB',
+                    total=total_mb,
+                    unit="MB",
                     unit_scale=True,
                     unit_divisor=1024,
-                    leave=False,  # Force horizontal display
-                    position=0,  # Lock the position
-                    dynamic_ncols=True,  # Adapt to terminal width
-                    bar_format='{l_bar}{bar}| {n:.2f}/{total:.2f}MB [{elapsed}<{remaining}, {rate_fmt}]',
+                    leave=False,
+                    position=0,
+                    dynamic_ncols=True,
+                    bar_format="{l_bar}{bar}| {n:.2f}/{total:.2f}MB [{elapsed}<{remaining}, {rate_fmt}]" if total_mb else "{desc}: {n:.2f}MB [{elapsed}, {rate_fmt}]",
                     desc=f"Downloading {os.path.basename(url)}",
-                    ascii=False
+                    ascii=False,
                 )
-                downloaded = 0
-                async for chunk in response.aiter_bytes(chunk_size=CHUNK_SIZE):
-                    content += chunk
-                    downloaded += len(chunk)
-                    progress_bar.update(len(chunk) / (1024 * 1024))  # Update in MB
-                progress_bar.close()
 
-                # Check if the content is gzip-compressed
-                if response.headers.get("Content-Encoding") == "gzip":
-                    content = gzip.decompress(content)
+                # Select optimal chunk size based on file size
+                if total_size > 0:
+                    if total_size > 500 * 1024 * 1024:  # > 500MB
+                        chunk_size = CHUNK_SIZE_LARGE
+                    elif total_size > 100 * 1024 * 1024:  # > 100MB
+                        chunk_size = CHUNK_SIZE_MEDIUM
+                    else:  # <= 100MB
+                        chunk_size = CHUNK_SIZE_SMALL
+                else:
+                    # Unknown size, use default
+                    chunk_size = CHUNK_SIZE_DEFAULT
 
-            if path:
-                if isinstance(path, str):
-                    path = Path(path)
-                if path.is_dir():
-                    path = path / os.path.basename(url)
+                # Always stream to temporary file
+                try:
+                    with open(temp_file, "wb") as f:
+                        # For large files, update progress less frequently to reduce overhead
+                        update_threshold = 1.0 if total_size > 500 * 1024 * 1024 else 0.1  # MB
+                        accumulated_mb = 0.0
+                        
+                        async for chunk in response.aiter_bytes(chunk_size=chunk_size):
+                            f.write(chunk)
+                            chunk_mb = len(chunk) / (1024 * 1024)
+                            accumulated_mb += chunk_mb
+                            
+                            # Update progress bar only when threshold is reached
+                            if accumulated_mb >= update_threshold:
+                                progress_bar.update(accumulated_mb)
+                                accumulated_mb = 0.0
+                        
+                        # Update any remaining progress
+                        if accumulated_mb > 0:
+                            progress_bar.update(accumulated_mb)
 
-            return save_or_return_content(content, path)
+                finally:
+                    progress_bar.close()
+
+                # Handle the result based on whether path was provided
+                if path is not None:
+                    # Atomic move to final destination
+                    final_path = Path(path)
+                    if final_path.is_dir():
+                        final_path = final_path / os.path.basename(url)
+
+                    # Ensure parent directory exists
+                    final_path.parent.mkdir(parents=True, exist_ok=True)
+
+                    # Atomic move from temp to final location
+                    shutil.move(str(temp_file), str(final_path))
+                    return None
+                else:
+                    # Return content for in-memory use
+                    if as_text:
+                        with open(temp_file, 'r', encoding='utf-8') as f:
+                            content = f.read()
+                    else:
+                        with open(temp_file, 'rb') as f:
+                            content = f.read()
+                    return content
+
+    finally:
+        # Clean up temporary directory
+        try:
+            shutil.rmtree(temp_dir, ignore_errors=True)
+        except Exception as e:
+            logger.warning(f"Failed to clean up temporary directory {temp_dir}: {e}")
 
 
 def download_json(data_url: str) -> dict:
@@ -597,8 +739,8 @@ def download_text_between_tags(url: str, tag: str):
     :param tag: The tag to extract the content from
 
     """
-    tag_start = f'<{tag}>'
-    tag_end = f'</{tag}>'
+    tag_start = f"<{tag}>"
+    tag_end = f"</{tag}>"
     is_header = False
     content = ""
 
@@ -616,19 +758,23 @@ def download_text_between_tags(url: str, tag: str):
 
                 # If within header lines, add to header_content
                 elif is_header:
-                    content += line + '\n'  # Add a newline to preserve original line breaks
+                    content += (
+                            line + "\n"
+                    )  # Add a newline to preserve original line breaks
     return content
 
 
-
-
 logger = logging.getLogger(__name__)
 
 
-@retry(on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
-async def download_bulk_data(url: str,
-                             data_directory: Path = get_edgar_data_directory(),
-                             client: Optional[AsyncClient] = None) -> Path:
+@retry(
+    on=RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial
+)
+async def download_bulk_data(
+        url: str,
+        data_directory: Path = get_edgar_data_directory(),
+        client: Optional[AsyncClient] = None,
+) -> Path:
     """
     Download and extract bulk data from zip or tar.gz archives
 
@@ -653,7 +799,7 @@ async def download_bulk_data(url: str,
     if not filename:
         raise ValueError("Invalid URL - cannot extract filename")
 
-    local_dir = filename.split('.')[0]
+    local_dir = filename.split(".")[0]
     download_path = data_directory / local_dir
     download_filename = download_path / filename
 
@@ -661,7 +807,7 @@ async def download_bulk_data(url: str,
         # Create the directory with parents=True and exist_ok=True to avoid race conditions
         download_path.mkdir(parents=True, exist_ok=True)
 
-        # Download the file
+        # Download the file (don't decompress archives)
         try:
             await stream_file(url, client=client, path=download_path)
         except Exception as e:
@@ -670,25 +816,49 @@ async def download_bulk_data(url: str,
         # Extract based on file extension
         try:
             if filename.endswith(".zip"):
-                with zipfile.ZipFile(download_filename, 'r') as z:
-                    z.extractall(download_path)
+                with zipfile.ZipFile(download_filename, "r") as z:
+                    # Calculate total size for progress bar
+                    total_size = sum(info.file_size for info in z.filelist)
+                    extracted_size = 0
+
+                    with tqdm(
+                            total=total_size, unit="B", unit_scale=True, desc="Extracting"
+                    ) as pbar:
+                        for info in z.filelist:
+                            z.extract(info, download_path)
+                            extracted_size += info.file_size
+                            pbar.update(info.file_size)
+
             elif any(filename.endswith(ext) for ext in (".tar.gz", ".tgz")):
-                with tarfile.open(download_filename, 'r:gz') as tar:
+                with tarfile.open(download_filename, "r:gz") as tar:
                     # Security check for tar files to prevent path traversal
                     def is_within_directory(directory: Path, target: Path) -> bool:
                         try:
-                            return os.path.commonpath([directory, target]) == str(directory)
+                            return os.path.commonpath([directory, target]) == str(
+                                directory
+                            )
                         except ValueError:
                             return False
 
-                    def safe_extract(tar: tarfile.TarFile, path: str) -> None:
-                        for member in tar.getmembers():
-                            member_path = os.path.join(path, member.name)
-                            if not is_within_directory(Path(path), Path(member_path)):
-                                raise ValueError(f"Attempted path traversal in tar file: {member.name}")
-                        tar.extractall(path)
+                    members = tar.getmembers()
+                    total_size = sum(member.size for member in members)
 
-                    safe_extract(tar, str(download_path))
+                    with tqdm(
+                            total=total_size, unit="B", unit_scale=True, desc="Extracting"
+                    ) as pbar:
+                        for member in members:
+                            # Check for path traversal
+                            member_path = os.path.join(str(download_path), member.name)
+                            if not is_within_directory(
+                                    Path(str(download_path)), Path(member_path)
+                            ):
+                                raise ValueError(
+                                    f"Attempted path traversal in tar file: {member.name}"
+                                )
+
+                            # Extract file and update progress
+                            tar.extract(member, str(download_path))
+                            pbar.update(member.size)
             else:
                 raise ValueError(f"Unsupported file format: {filename}")
 
@@ -701,7 +871,9 @@ async def download_bulk_data(url: str,
                 if download_filename.exists():
                     download_filename.unlink()
             except Exception as e:
-                logger.warning(f"Failed to delete archive file {download_filename}: {e}")
+                logger.warning(
+                    f"Failed to delete archive file {download_filename}: {e}"
+                )
 
         return download_path
 
@@ -715,8 +887,7 @@ async def download_bulk_data(url: str,
         raise
 
 
-
-def download_datafile(data_url: str, local_directory:Path=None) -> Path:
+def download_datafile(data_url: str, local_directory: Path = None) -> Path:
     """Download a file to the local storage directory"""
     filename = os.path.basename(data_url)
     # Create the directory if it doesn't exist
